# CheapBulletWhizz release 2 by David Bowland
# ./addons/eventscripts/cheapbulletwhizz/cheapbulletwhizz.py

# >>> To configure this addon please see cheapbulletwhizz.cfg (created when the addon is first loaded) <<<

"""
Provides bullet whizz sounds. Requires only ES 2.0.0.248+
"""


import cfglib
import es
import math
import os.path
import random

import psyco
psyco.full()


info = es.AddonInfo()
info.name     = 'CheapBulletWhizz'
info.version  = '2'
info.basename = 'cheapbulletwhizz'
info.url      = 'http://addons.eventscripts.com/addons/view/' + info.basename
info.author   = 'SuperDave'


###


config = cfglib.AddonCFG(es.getAddonPath(info.basename) + '/' + info.basename + '.cfg')

config.text(info.name + ' release %s options' % info.version)
config.text('./addons/eventscripts/%(basename)s/%(basename)s.cfg' % {'basename':info.basename})
config.text(info.url)
config.text('')
config.text('Load this addon with: es_load ' + info.basename)
config.text('')
config.text('This addon will turn on eventscripts_noisy. More information on eventscripts_noisy can be found here:')
config.text('http://www.eventscripts.com/pages/Eventscripts_noisy')
config.text('')
config.text('Thanks to Juba_PornBorn for this idea! Juba\'s bullet whizz thread can be found here:')
config.text('http://forums.mattie.info/cs/forums/viewtopic.php?p=185915#185915')
config.text('')
config.text('Juba got the idea from a plugin called Bullet Whizz which can be found here:')
config.text('http://forums.alliedmods.net/showthread.php?t=29852')
config.text('\n')

config.text('***** General options *****')
cvar_distance_min  = config.cvar('cheapbulletwhizz_distance_min',  50,  'Distance from a bullet\'s trajectory at which players hear whizz sound at maximum volume')
cvar_distance_max  = config.cvar('cheapbulletwhizz_distance_max',  300, 'Maximum distance a bullet\'s trajectory can be to hear a whizz sound')
cvar_opponentsonly = config.cvar('cheapbulletwhizz_opponentsonly', 0,   '0 = players hear the whizz from all bullets, 1 = players hear the whizz from bullets of opponents only')
config.text('\n')

config.text('***** Sound options *****')
config.text('')
config.text('Use the following server command to add bullet whizz sounds:')
config.text('')
config.text('cheapbulletwhizz_addsound <"sound relative to ./sound/">')
config.text('')
config.text('Sounds are marked for download automatically.')
config.text('\n')

config.text('Place your cheapbulletwhizz_addsound commands below:')
config.command('cheapbulletwhizz_addsound')
config.text('\n')

config.text('Uncomment the following default options if you wish to use them:')
config.text('')
config.text('cheapbulletwhizz_addsound "ambient/animal/flies1.wav"')
config.text('cheapbulletwhizz_addsound "ambient/animal/flies2.wav"')
config.text('cheapbulletwhizz_addsound "ambient/animal/flies3.wav"')
config.text('cheapbulletwhizz_addsound "ambient/animal/flies4.wav"')
config.text('cheapbulletwhizz_addsound "ambient/animal/flies5.wav"')
config.text('\n')

config.text('The following server command can be used to remove bullet whizz sounds:')
config.text('')
config.text('cheapbulletwhizz_removesound <"sound relative to ./sound/">')
config.text('')
config.text('Due to the fact bullet whizz sounds can be added and removed above, this command will not be used by most users')
config.text('\n')

config.text('***** Event documentation *****')
config.text('')
config.text('The following event will fire while BulletWhizz is loaded:')
config.text('')
config.text('bullet_whizz')
config.text('- Fires when a player hears a bullet whizz sound')
config.text('   event_var(userid)   - Userid of the player who heard the sound')
config.text('   event_var(distance) - Distance from the player the bullet traveled')
config.text('   event_var(attacker) - Userid of the player who fired the bullet')
config.text('')
config.text('This event also provides event_var(es_steamid), event_var(es_username), and all the other event_vars that accompany')
config.text('event_var(userid). More information can be found here: http://www.eventscripts.com/pages/Extended_event_variables')

config.write()

sound_path  = es.ServerVar('eventscripts_gamedir') + '/sound/'


###


class BulletWhizzManager(object):
   last_sound = ''

   def __init__(self, userid):
      self.userid = userid

      self.reset()

   def reset(self):
      """Resets bullet stats"""
      self.last_fire    = ()

      self.hurt_players = set()

   def weaponFire(self):
      """Records the origin of the bullet"""
      location = list(es.getplayerlocation(self.userid))
      location[2] += 45 if es.getplayerprop(self.userid, 'CBasePlayer.localdata.m_Local.m_bDucked') % 2 else 60

      self.last_fire    = tuple(location)

      self.hurt_players = set((self.userid,))

   def hurt(self, userid):
      """Marks the player as hurt by the bullet"""
      self.hurt_players.add(userid)

   def bulletImpact(self, impact_location):
      """
      Calculates the slope and distance of the shot
      Loops through a filtered list of players and retrieves each player's distance to the bullet's trajectory
      Plays the whizz sound and fires bullet_whizz
      """
      if not self.last_fire: return

      impact_slopes   = self._getSlopes(self.last_fire, impact_location)
      impact_distance = sum((self.last_fire[x] - impact_location[x]) ** 2 for x in xrange(3)) ** 0.5

      distance_min = float(cvar_distance_min)
      distance_max = float(cvar_distance_max)

      if int(cvar_opponentsonly):
         team    = es.getplayerteam(self.userid)
         userids = filter(lambda x: es.getplayerteam(x) == 5 - team, es.getUseridList())

      else:
         userids = es.getUseridList()

      for userid in filter(lambda x: not es.getplayerprop(x, 'CBasePlayer.pl.deadflag'), set(userids).difference(self.hurt_players)):
         player_location     = list(es.getplayerlocation(userid))
         player_location[2] += 45 if es.getplayerprop(userid, 'CBasePlayer.localdata.m_Local.m_bDucked') % 2 else 60

         player_distance = sum((player_location[x] - self.last_fire[x]) ** 2 for x in xrange(3)) ** 0.5
         player_impact   = sum((player_location[x] - impact_location[x]) ** 2 for x in xrange(3)) ** 0.5

         angles_to_player = self._getAngles(impact_slopes, self._getSlopes(self.last_fire, player_location))
         distance_to_shot = sum((math.sin(angles_to_player[x]) * player_distance) ** 2 for x in xrange(3)) ** 0.5

         max_dist = (impact_distance ** 2 + distance_to_shot ** 2) ** 0.5
         if max_dist <= player_distance or max_dist <= player_impact:
            continue

         volume = 1.0 - (distance_to_shot - distance_min) / distance_max
         if volume > 0:
            if sounds:
               es.playsound(userid, self._getSound(userid), volume if volume < 1 else 1)

            es.event('initialize', 'bullet_whizz')
            es.event('setint',     'bullet_whizz', 'userid',   userid)
            es.event('setfloat',   'bullet_whizz', 'distance', distance_to_shot)
            es.event('setint',     'bullet_whizz', 'attacker', self.userid)
            es.event('fire',       'bullet_whizz')

   def _getAngles(self, slopes1, slopes2):
      """Returns the angles between two slopes in a 3-element tuple"""
      return tuple(math.atan(abs(slopes1[x] - slopes2[x]) / (1 + slopes1[x] * slopes2[x])) for x in xrange(3))

   def _getSlopes(self, point1, point2):
      """Returns the slope between two points in a 3-element tuple"""
      return tuple(self.__getSingleSlope(point1[x], point1[y], point2[x], point2[y]) for x, y in ((0, 1), (1, 2), (0, 2)))

   def __getSingleSlope(self, x1, y1, x2, y2):
      """Returns the slope between two points"""
      return ((y2 - y1) / (x2 - x1)) if x2 - x1 else 0

   def _getSound(self, userid):
      """Returns a whizz sound avoiding the last sound if applicable"""
      choices = filter(lambda x: x <> self.last_sound, sounds) if self.last_sound else ()
      current_sound = self.last_sound = random.choice(choices if choices else sounds)

      return current_sound


players = {}


def getPlayer(userid):
   """Returns the BulletWhizzManager instance of the player"""
   userid = int(userid)
   if userid not in players:
      players[userid] = BulletWhizzManager(userid)

   return players[userid]


def removePlayer(userid):
   """Removes the BulletWhizzManager instance of the player"""
   userid = int(userid)
   if userid in players:
      del players[userid]


###


def load():
   """
   Ensures critical server variables are created by the config
   Calls mark_sounds_for_download
   Sets eventscripts_noisy to 1
   """
   config.execute()

   load_events()
   mark_sounds_for_download()

   es.doblock('corelib/noisy_on')


def es_map_start(event_var):
   """
   Calls load_events
   Calls mark_sounds_for_download
   Clears ditionary of players and last sounds played
   """
   load_events()
   mark_sounds_for_download()

   players.clear()


def player_spawn(event_var):
   """Resets the player's bullet stats"""
   getPlayer(event_var['userid']).reset()


def player_hurt(event_var):
   """Marks the player as hurt by the attacker"""
   attacker = int(event_var['attacker'])
   if attacker:
      getPlayer(attacker).hurt(int(event_var['userid']))


def weapon_fire(event_var):
   """Stores the player's position of weapon fire"""
   getPlayer(event_var['userid']).weaponFire()


def bullet_impact(event_var):
   """Initiates whizz sounds"""
   getPlayer(event_var['userid']).bulletImpact(tuple(float(event_var[x]) for x in ('x', 'y', 'z')))


def player_disconnect(event_var):
   """Removes the disconnecting player's BulletWhizzManager instance"""
   removePlayer(event_var['userid'])


def unload():
   """Reverts eventscripts_noisy"""
   es.doblock('corelib/noisy_off')

   for cvar in config.getCvars():
      es.ServerVar(cvar).set(0)


###


sounds = []


def addsound_cmd():
   """
   cheapbulletwhizz_addsound <"sound relative to ./sound/">
   Adds a bullet whizz sound
   """
   if es.getargc() == 2:
      sounds.append(es.getargv(1))

   else:
      es.dbgmsg(0, 'Syntax: cheapbulletwhizz_addsound <\"sound relative to ./sound/\">')
if not es.exists('command', 'cheapbulletwhizz_addsound'):  
   es.regcmd('cheapbulletwhizz_addsound', 'cheapbulletwhizz/addsound_cmd', 'cheapbulletwhizz_addsound <\"sound relative to ./sound/\">\nAdds a bullet whizz sound')


def removesound_cmd():
   """
   cheapbulletwhizz_removesound <"sound relative to ./sound/">
   Removes a bullet whizz sound
   """
   if es.getargc() == 2:
      sound = es.getargv(1)
      if sound in sounds: 
         sounds.remove(sound)

   else:
      es.dbgmsg(0, 'Syntax: cheapbulletwhizz_removesound <\"sound relative to ./sound/\">')
if not es.exists('command', 'cheapbulletwhizz_removesound'):  
   es.regcmd('cheapbulletwhizz_removesound', 'cheapbulletwhizz/removesound_cmd', 'cheapbulletwhizz_removesound <\"sound relative to ./sound/\">\nRemoves a bullet whizz sound')


###


def mark_sounds_for_download():
   """Marks whizz sounds for download"""
   for sound in filter(lambda x: os.path.isfile(sound_path + x), sounds):
      es.stringtable('downloadables', 'sound/' + sound)


def load_events():
   """Loads the .res file for the events"""
   events = {'bullet_whizz':{'userid':'short', 'distance':'float', 'attacker':'short'}}
   if os.path.isfile(es.getAddonPath('cheapbulletwhizz') + '/cheapbulletwhizz.res'):
      es.loadevents('declare', 'addons/eventscripts/cheapbulletwhizz/cheapbulletwhizz.res')

   else:
      if not os.path.isfile(es.getAddonPath('cheapbulletwhizz') + 'es_cheapbulletwhizz_events_db.txt'):
         es.keygroupcreate('cheapbulletwhizz_events')
         for event_name in events:
            es.keycreate('cheapbulletwhizz_events', event_name)
            for event_var in events[event_name]:
               es.keysetvalue('cheapbulletwhizz_events', event_name, event_var, events[event_name][event_var])
         es.keygroupsave('cheapbulletwhizz_events', '|cheapbulletwhizz')
         es.keygroupdelete('cheapbulletwhizz_events')
      es.loadevents('declare', 'addons/eventscripts/cheapbulletwhizz/es_cheapbulletwhizz_events_db.txt')
      es.dbgmsg(0, 'CheapBulletWhizz: Unable to load cheapbulletwhizz.res! Please ensure it is in the ./cheapbulletwhizz/ directory.')